import os
from glob import glob
import tarfile
import shutil
from tempfile import mkstemp
import subprocess
import sys
import time
from pynwp.migration import IncrementalUpdates, SystemUtil
from pynwp.Utils import TempTree


class ConfigMover:
    param_names = ["mode", "perm", "user", "group"]

    def read_rules(self, filelist):
        rules = {}
        with open(filelist, mode='r') as f:
            for line in f:
                stripped = line.strip()
                if stripped.startswith('#') or stripped == '':
                    continue
                new_rule = self.create_rule(stripped)
                if new_rule is None:
                    continue
                (new_key, new_value) = new_rule
                if new_key in rules:
                    raise ConfigMover.FileSchemaError('Duplicating rules')
                rules[new_key] = new_value
        return rules

    def create_rule(self, stripped):
        words = stripped.split()
        path = words[0]
        params = {name: value
                  for name, value in zip(self.param_names, words[1:])}
        if 'mode' not in params:
            return None
        return (path, params)

    @staticmethod
    def read_schema_exclude(path, rootfs):
        path_list = []
        try:
            if path:
                with open(path) as f:
                    path_list = [line.strip() for line in f]
        except IOError:
            return path_list

        path_list_abs = [os.path.join(rootfs, item.strip('/')) for item in path_list]

        return path_list_abs

    def find_files(self, rootfs, filelist, filter_modes=None, exclude=None):
        found = {}

        rules = self.read_rules(filelist)
        ignored = self.read_schema_exclude(path=exclude, rootfs=rootfs)

        for path, params in rules.items():
            if filter_modes is not None:
                if params['mode'] not in filter_modes:
                    continue

            item_path = os.path.join(rootfs, path.strip('/'))
            for expanded in glob(item_path):
                if expanded not in ignored and os.path.isfile(expanded):
                    if expanded in found:
                        raise ConfigMover.FileSchemaError(
                            'Multiple file matching')
                    found[expanded] = params

        return found

    def op_save_cfg(self, root, schema, tar_path, version, schema_exclude=None, **_):
        files = self.find_files(rootfs=root, filelist=schema, filter_modes=('file'), exclude=schema_exclude)
        if version != None:
            pax_headers = {'version': str(version)}
        else:
            pax_headers = None
        with tarfile.open(name=tar_path, mode='w:',
                          pax_headers=pax_headers,
                          format=tarfile.PAX_FORMAT) as f:
            for item in files.keys():
                f.add(name=item,
                      arcname=os.path.relpath(item, root), recursive=False)

    @staticmethod
    def op_get_bundle_metadata(tar_path, datum):
        try:
            with tarfile.open(name=tar_path) as f:
                try:
                    return 0, f.pax_headers[datum]
                except:
                    return 2, "Can't obtain datum '{}'".format(datum)
        except Exception as e:
            return 1, "Can't open '{}': {}.".format(tar_path, e)

    def op_save_log(self, root, tar_path, **_):
        src = os.path.join(root, self.default_log_dir())

        with tarfile.open(name=tar_path, mode='w:gz') as f:
            for dirpath, _, files_rel in os.walk(src):
                files_abs = (os.path.join(dirpath, item) for item in files_rel)
                for item in files_abs:
                    f.add(name=item,
                          arcname=os.path.relpath(item, root),
                          recursive=False)

    def op_save_db(self, root, bkp_path=None, db_path=None, tbl_filter=None, **_):
        if not db_path:
            db_path = os.path.join(root, self.default_db_path())
        if not bkp_path:
            bkp_path = os.path.join(root, self.default_dbbackup_path())

        self.sqlite_call(db_path, '.backup {0}'.format(bkp_path))

        all_tables = self.db_tables(db_path)
        if tbl_filter:
            for name in filter(lambda x: not tbl_filter(x), all_tables):
                self.sqlite_call(bkp_path,
                                 'drop table if exists {0}'.format(name))

    def op_dbrestore(self, root, db_path=None, bkp_path=None, **_):
        if not db_path:
            db_path = os.path.join(root, self.default_db_path())
        if not bkp_path:
            bkp_path = os.path.join(root, self.default_dbbackup_path())

        # sqlite .restore command deletes tables in dst if they are not present in src,
        # so we iterate over tables in src and copy their content individually
        target_tables = self.db_tables(db_path)
        src_tables = self.db_tables(bkp_path)
        for name in (tbl for tbl in src_tables if tbl in target_tables):
            command = '; '.join([
                'delete from {0}',
                'attach "{1}" as bkp',
                'insert into {0} select * from bkp.{0}'
            ]).format(name, bkp_path)
            self.sqlite_call(db_path, command)

    def db_tables(self, db_path):
        output = self.sqlite_call(db_path, '.tables')
        return output.stdout.split()

    @staticmethod
    def default_dbbackup_path():
        return 'home/spectracom/config/lafayette.bkp'

    @staticmethod
    def default_db_path():
        return 'srv/www2/app/webroot/lafayette'

    @staticmethod
    def default_log_dir():
        return 'home/spectracom/log'

    @staticmethod
    def log_db_tables():
        return {'log_events', 'log_sys_mons', 'log_eth_mons', 'log_gps_statuses', 'log_references', 'log_oscillators', 'log_ntp_stats'}

    @staticmethod
    def default_cfg_dir():
        return 'home/spectracom/default'

    @staticmethod
    def default_ntpstat_dir():
        return 'home/spectracom/log/ntpstats'

    @staticmethod
    def clear_dir(target):
        if os.path.exists(target):
            for name in os.listdir(target):
                abs_path = os.path.join(target, name)
                if os.path.isfile(abs_path):
                    os.remove(abs_path)

    @staticmethod
    def tar_extract(root, tar_path):
        with tarfile.open(tar_path, 'r') as tar:
            tar.extractall(root)

    def op_load_cfg(self, root, tar_path, schema, patch_args=None, **_):
        with TempTree() as tmp_root, TempTree() as tmp_files:
            self.tar_extract(root=tmp_root, tar_path=tar_path)

            if patch_args:
                patcher = IncrementalUpdates(current_version=patch_args.source_version,
                                             target_version=patch_args.target_version,
                                             patches_path=patch_args.patches_path,
                                             rootfs=tmp_root,
                                             rootfs_src=tmp_root,
                                             rootfs_dflt=os.path.join(root, self.default_cfg_dir()),
                                             board_type=SystemUtil.board_type())
                patcher.execute_upgrade_patches()

            self.apply_schema_permissions(rootfs=tmp_root, schema=schema)

            sanitized_bundle = os.path.join(tmp_files, 'bundle')
            self.op_save_cfg(root=tmp_root,
                             schema=schema,
                             tar_path=sanitized_bundle,
                             version=None)

            self.tar_extract(root=root, tar_path=sanitized_bundle)

    def apply_schema_permissions(self, rootfs, schema):
        files = self.find_files(rootfs=rootfs,
                                filelist=schema,
                                filter_modes=('file'))
        for path, params in files.items():
            os.chmod(path, int(params['perm'], 8))
            shutil.chown(path,
                         self.try_convert_digits(params['user']),
                         self.try_convert_digits(params['group']))

    @staticmethod
    def try_convert_digits(x):
        return int(x) if x.isdigit() else x

    @staticmethod
    def _get_file_owner(path):
        stat = os.stat(path, follow_symlinks=False)
        return {'uid': stat.st_uid, 'gid': stat.st_gid}

    def op_restore_default(self, root, **_):
        default_dir = os.path.join(root, self.default_cfg_dir())
        for node_src, dirs, files in os.walk(default_dir):
            node_dst = node_src.replace(default_dir, root)
            if not os.path.exists(node_dst):
                os.mkdir(node_dst)
                shutil.copymode(src=node_src, dst=node_dst)

            for item in files:

                src_file = os.path.join(node_src, item)
                dst_file = os.path.join(node_dst, item)
                if os.path.lexists(dst_file):
                    os.remove(dst_file)
                shutil.copy(src=src_file, dst=dst_file, follow_symlinks=False)
                owner = self._get_file_owner(src_file)
                os.chown(dst_file, owner['uid'], owner['gid'], follow_symlinks=False)

            if not dirs and not files:
                files_to_delete = (item
                                   for item in os.listdir(node_dst)
                                   if os.path.isfile(os.path.join(node_dst,
                                                                  item)))
                for item in files_to_delete:
                    dst_file = os.path.join(node_dst, item)
                    os.remove(dst_file)

        db_backup = mkstemp()[1]
        self.op_save_db(tbl_filter=lambda name: name in self.log_db_tables(),
                        bkp_path=db_backup,
                        root=root)
        self.shell_call("sql_create.sh --force", shell=True)
        self.op_dbrestore(bkp_path=db_backup,
                          root=root)
        os.remove(db_backup)

    def op_clear_ntpstats(self, root, **_):
        log_dir = os.path.join(root, self.default_ntpstat_dir())
        self.clear_dir(log_dir)

    class FileSchemaError(ValueError):
        pass

    @staticmethod
    def shell_call(cmd, retry=None, timeout=None, **run_args):
        attempts = range(retry if retry else 1)
        for i in attempts:
            ret = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True, **run_args)
            if ret.returncode == 0:
                break
            # in case of error try again
            if timeout and i != attempts[-1]:
                # sleep if it's not the last one
                time.sleep(timeout)

        if ret.returncode != 0:
            # if last call still failed, it's a fatal error
            print(ret.stderr, file=sys.stderr, end='')
            sys.exit(ret.returncode)
        return ret

    @staticmethod
    def sqlite_call(db_path, cmd, retry=20, timeout=0.2, **run_args):
        return ConfigMover.shell_call(['sqlite3', db_path, cmd], retry, timeout, **run_args)
