#!/usr/bin/python3

import os
import re
import sys


class SystemdTarget:
    def __init__(self, name, units):
        self.name = name
        self.units = units


class SystemdPreset:
    preset_modes = {'enable', 'disable'}

    def __init__(self, root='/'):
        self.root = root
        self.dir_preset = os.path.join(root, 'usr/lib/systemd/system-preset')

    def active(self):
        return os.path.isdir(self.dir_preset)

    @staticmethod
    def get_default_preset_files(dir_preset):
        # make preset list for specified directory in alphabetical order
        try:
            entries = sorted(os.listdir(dir_preset))
        except OSError:
            return

        # arrange them in targets dictionary
        # each entry is a list of corresponding files
        # filenames are sorted before
        # filename starts with digits, then hyphen, then target name, then '.preset' extension
        re_file = re.compile(r'^\d+-(?:[a-z]+)-(.*)\.preset')
        files = {}
        for entry in entries:
            search_result = re_file.search(entry)
            if search_result is not None:
                target = search_result.group(1)
                files.setdefault(target, [])
                files[target].append(entry)
        return files

    def get_preset_units(self):
        units = {mode: {} for mode in self.preset_modes}

        files = self.get_default_preset_files(self.dir_preset)
        for file in sorted(files, reverse=True):
            with open(file, 'r') as fp:
                more_units = self.read_preset(fp)
                for mode, unit_list in more_units.items():
                    units[mode].update(unit_list)
        return units

    @staticmethod
    def read_preset(fp):
        target_units = {mode: set() for mode in SystemdPreset.preset_modes}
        all_units = set()

        for line in fp:
            # match only non-empty non-commented lines
            stripped = line.strip()
            if stripped.startswith('#') or stripped == '':
                continue

            # match only lines containing two words, first is enable or disable
            words = stripped.split()
            if len(words) != 2:
                continue
            (mode, unit) = words
            if mode not in SystemdPreset.preset_modes:
                continue
            if unit in all_units:
                continue
            all_units.add(unit)

            target_units[mode].add(unit)

        return target_units

    def GetDisabledServices(self):
        target_presets = SystemdPreset.get_default_preset_files(self.dir_preset)

        # read preset files compose lists of units to be enabled
        # if units was mentioned once, later occurences in presets of lesser priority don't count
        target_units = []
        for target_name, preset_list in target_presets.items():
            for preset in preset_list:
                with open(os.path.join(self.dir_preset, preset), 'r') as fp:
                    target_units.append(SystemdTarget(name=target_name,
                                                      units=self.read_preset(fp)))

        # iterate over targets, check if vendor enabled services are actually enabled
        target_disabled = {}
        for target in target_units:
            target_disabled.setdefault(target.name, [])
            dir_target = self.GetTargetDir(target.name)
            for unit in target.units['enable']:
                if os.path.islink(os.path.join(dir_target, unit)) is False:
                    target_disabled[target.name].append(unit)

        return target_disabled

    def GetTargetDir(self, target_name):
        return os.path.join(self.root, 'etc/systemd/system', '{}.target.wants'.format(target_name))

    def DisableServices(self, target_disable):
        # iterate over dictionary
        # keys are targets, values are arrays of units
        for target, units in target_disable.items():
            dir_target = self.GetTargetDir(target)
            for unit in units:
                try:
                    os.remove(os.path.join(dir_target, unit))
                except OSError:
                    pass


if __name__ == '__main__':
    disabled_services = SystemdPreset(sys.argv[1]).GetDisabledServices()
